/* Basic inflate procedure, working on files. */

/* Copyright (C) 2009 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>

#include "config.h"
#include "inflate.h"
#include "huffman.h"
#include "fixed-code.h"

#define HLIT_MAX 286
#define HDIST_MAX 30
#define HCLEN_MAX 19
#define CLEN_MAX (HLIT_MAX + HDIST_MAX)

static void
PRINTF_STYLE(1, 2)  
werror(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
}

/* By RFC 1951, we treat a byte sequence as a sequence of bits, where
 * the least signifocant bit is the first bit in each byte. */
struct input {
  FILE *f;
  unsigned buffer;
  unsigned bits;
};

static void
input_init (struct input *self, FILE *f)
{
  self->f = f;
  self->buffer = self->bits = 0;
}

static int
input_bits(struct input *self, unsigned bits)
{
  int out;

  assert(bits > 0);
  assert(bits < sizeof(self->buffer) * CHAR_BIT);
  assert(self->bits < 8);

  while (self->bits < bits)
    {
      int c = getc(self->f);
      if (c < 0)
	return -1;
      self->buffer += (c << self->bits);
      self->bits += 8;
    }
  out = self->buffer & ((1 << bits) - 1);
  self->buffer >>= bits;
  self->bits -= bits;

  return out;
}

static int
input_decode(struct input *self, const struct huffman_node *tree)
{
  struct huffman_ctx ctx;
  unsigned value;
  int res;

  huffman_init (&ctx, tree);
  do
    {
      int bit = input_bits(self, 1);
      if (bit < 0)
	return -1;
      res = huffman_decode_bit(&ctx, bit, &value);
    }
  while (!res);

  return res > 0 ? (int) value : res;
}

enum length_lit_type
  {
    LENGTH_LIT_ERROR,
    LENGTH_LIT_END,
    LENGTH_LIT_LITERAL,
    LENGTH_LIT_LENGTH
  };

/* Returns -1 for error, 0 for end of block, 1 for a literal, and 2 for a length. */
static enum length_lit_type
input_length_lit(struct input *self, const struct huffman_node *tree,
		 unsigned *value)
{
  int length;

  length = input_decode(self, tree);

  /* NOTE: The invalid values 286 and 287 can occur only when using
     the fixed coding, which is defined to include them. */
  if (length < 0 || length >= HLIT_MAX)
    return LENGTH_LIT_ERROR;

  else if (length < 256)
    {
      *value = length;
      return LENGTH_LIT_LITERAL;
    }
  else if (length == 256)
    return LENGTH_LIT_END;
  else
    {
      int bits;
      static const unsigned base_table[HLIT_MAX - 257] =
	{ 3, 4, 5, 6, 7, 8, 9, 10,
	  11, 13, 15, 17,
	  19, 23, 27, 31,
	  35, 43, 51, 59,
	  67, 83, 99, 115,
	  131, 163, 195, 227,
	  258
	};      

      if (length >= 265 && length < 285)
	{
	  unsigned nbits = (length - 261) / 4;
	  assert(nbits > 0);

	  bits = input_bits (self, nbits);
	  if (bits < 0)
	    return LENGTH_LIT_ERROR;
	}
      else
	bits = 0;

      *value = base_table[length - 257] + bits;
      return LENGTH_LIT_LENGTH;
    }
}

static int
input_distance(struct input *self, const struct huffman_node *tree)
{
  int distance;

  distance = input_decode(self, tree);

  if (distance < 0)
    return -1;

  assert(distance < HDIST_MAX);

  if (distance >= 4)
    {
      unsigned nbits = (distance - 2) / 2;
      int bits = input_bits(self, nbits);
      if (bits < 0)
	return 0;

      distance = ((2 + (distance & 1)) << nbits) + bits;
    }
  return distance + 1;
}

static void
input_reset(struct input *self)
{
  self->bits = 0;
}

#define MAX_WINDOW 32768
struct output
{
  FILE *f;
  /* Circular output buffer. */
  unsigned start;
  unsigned size;
  unsigned char window[MAX_WINDOW];
};

static void
output_init (struct output *self, FILE *f)
{
  self->f = f;
  self->start = self->size = 0;
}

static int
output_char(struct output *self, int c)
{
  assert(self->size <= MAX_WINDOW);
  assert(self->start < MAX_WINDOW);

  if (self->size < MAX_WINDOW)
    {
      unsigned end = self->start + self->size;
      if (end > MAX_WINDOW)
	end -= MAX_WINDOW;
      self->window[end] = c;
      self->size++;
    }
  else
    {
      self->window[self->start++] = c;
      if (self->start == MAX_WINDOW)
	self->start = 0;
    }
  return (putc(c, self->f) >= 0);
}

static int
output_match(struct output *self, unsigned length, unsigned distance)
{
  unsigned match;

  assert(length > 0);
  assert(distance > 0);
  assert(distance <= MAX_WINDOW);

  if (distance > self->size)
    return 0;

  match = self->start + self->size - distance;
  if (match >= MAX_WINDOW)
    match -= MAX_WINDOW;

  /* Note that length > distance is allowed. */
  /* Simple but inefficient. */
  for (; length > 0; length--)
    {
      assert(match <= MAX_WINDOW);

      if (!output_char(self, self->window[match]))
	return 0;

      match++;
      if (match >= MAX_WINDOW)
	match -= MAX_WINDOW;
    }
  return 1;
}

static int
copy_stream (FILE *from, FILE *to, unsigned length, unsigned char *buf)
{
  return (fread(buf, length, 1, from) == 1
	  && fwrite(buf, length, 1, to) == 1);
}

static int
output_stream(struct output *self, unsigned length, FILE *f)
{
  assert(length > 0);
  assert(length <= 2*MAX_WINDOW);
  assert(self->size <= MAX_WINDOW);
  assert(self->start < MAX_WINDOW);

  /* FIXME: Seems to not update size properly. */
  if (length < MAX_WINDOW)
    {
      /* The new window will contain both old and new data. */
      if (self->start + length <= MAX_WINDOW)
	{
	  if (copy_stream(f, self->f, length, self->window + self->start))
	    return 0;
	}
      else
	{
	  unsigned atend = MAX_WINDOW - self->start;
	  if (!(copy_stream(f, self->f, atend, self->window + self->start)
		&& copy_stream(f, self->f, length - atend, self->window)))
	    return 0;
	}
      self->size += length;
      if (self->size > MAX_WINDOW)
	{
	  self->start += (self->size - MAX_WINDOW);
	  if (self->start > MAX_WINDOW)
	    self->start -= MAX_WINDOW;

	  self->size = MAX_WINDOW;
	}
    }
  else 
    {
      unsigned forgotten = length - MAX_WINDOW;
      if (forgotten > 0)
	if (!copy_stream(f, self->f, forgotten, self->window))
	  return 0;

      if (!copy_stream(f, self->f, MAX_WINDOW, self->window))
	return 0;

      self->size = MAX_WINDOW;
    }
  return 1;
}

/* Used for both code table (length 19), and the main decoding tables,
   of length (256 + 1 + 29 + 30) */

int
inflate_stream(FILE *in, FILE *out)
{
  struct input input;
  struct output output;
  int final;

  input_init(&input, in);
  output_init(&output, out);

  do
    {
      int header = input_bits(&input, 3);
      int type;

      if (header < 0)
	return 0;
      final = header & 1;
      type = header >> 1;

      werror("Block of type %d, final = %d\n", type, final);

      if (type == 3)
	{
	  werror("Unsupported block type.\n");
	  return 0;
	}
      else if (type == 0)
	{
	  /* No compression. */
	  int length;
	  int nlength;

	  input_reset(&input);
	  if ((length = input_bits(&input, 16)) < 0
	      || (nlength = input_bits(&input, 16)) < 0
	      || (length + nlength) != 0xffff)
	    {
	      werror("Bad header for uncompressed block.\n");
	      return 0;
	    }

	  if (length > 0)
	    output_stream(&output, length, input.f);
	}
      else
	{
	  struct huffman_node length_tree[HUFFMAN_SIZE(HLIT_MAX)];
	  struct huffman_node distance_tree[HUFFMAN_SIZE(HDIST_MAX)];
	  const struct huffman_node *length_active;
	  const struct huffman_node *distance_active;
	  
	  if (type == 1)
	    {
	      length_active = fixed_length_tree;
	      distance_active = fixed_distance_tree;
	    }
	  else
	    {
	      /* Dynamic table */
	      int hlit;
	      int hdist;
	      int hclen;

	      unsigned char clen[CLEN_MAX];
	      unsigned k;
	      struct huffman_node code_tree[HUFFMAN_SIZE(HCLEN_MAX)];

	      static const unsigned char permutation[HCLEN_MAX] =
		{ 16, 17, 18, 0, 8, 7, 9, 6, 10, 5,
		  11, 4, 12, 3, 13, 2, 14, 1, 15 };

	      assert(type == 2);

	      hlit = input_bits(&input, 5);
	      if (hlit < 0)
		return 0;
	      hlit += 257;
	      if (hlit > HLIT_MAX)
		return 0;

	      hdist = input_bits(&input, 5);
	      if (hdist < 0)
		return 0;
	      hdist += 1;
	      if (hdist > HDIST_MAX)
		return 0;

	      hclen = input_bits(&input, 4);
	      if (hclen < 0)
		return 0;
	      hclen += 4;

	      assert(hclen <= HCLEN_MAX);
		     
	      werror("hlit = %d, hdist = %d, hclen = %d\n",
		     hlit, hdist, hclen);

	      memset(clen, 0, sizeof(clen));
	      for (k = 0; k < hclen; k++)
		{
		  int bits = input_bits(&input, 3);
		  if (bits < 0)
		    return 0;

		  clen[permutation[k]] = bits;
		}

	      if (!huffman_tree(code_tree, HCLEN_MAX, clen))
		return 0;

	      for (k = 0; k < hlit + hdist; )
		{
		  int code = input_decode(&input, code_tree);
		  if (code < 0)
		    return 0;
		  if (code < 16)
		    clen[k++] = code;
		  else
		    {
		      unsigned value;
		      unsigned repeat;
		      unsigned nbits;
		      int bits;
		      
		      if (code == 16)
			{
			  if (k == 0)
			    return 0;

			  value = clen[k-1];
			  repeat = 3;
			  nbits = 2;
			}
		      else if (code == 17)
			{
			  value = 0;
			  repeat = 3;
			  nbits = 3;
			}
		      else
			{
			  assert(code == 18);

			  value = 0;
			  repeat = 11;
			  nbits = 7;
			}

		      bits = input_bits(&input, nbits);
		      if (bits < 0)
			return 0;
		      repeat += bits;
		      if (k + repeat > hlit + hdist)
			return 0;
		      memset(clen + k, value, repeat);
		      k+= repeat;
		    }
		}

	      if (!huffman_tree(length_tree, hlit, clen))
		return 0;
	      if (!huffman_tree(distance_tree, hdist, clen + hlit))
		return 0;

	      length_active = length_tree;
	      distance_active = distance_tree;
	    }
	  for (;;)
	    {
	      unsigned value;
	      switch (input_length_lit(&input, length_active, &value))
		{
		case LENGTH_LIT_ERROR:
		  return 0;
		case LENGTH_LIT_END:
		  /* Break out of for loop. */
		  goto next_block;
		case LENGTH_LIT_LITERAL:
		  if (!output_char(&output, value))
		    return 0;
		  break;
		case LENGTH_LIT_LENGTH:
		  {
		    int distance;

		    distance = input_distance(&input, distance_active);
		    if (distance < 0)
		      return 0;

		    if (!output_match(&output, value, distance))
		      return 0;

		    break;
		  }
		}
	    }
	next_block:
	  ;
	}
    }
  while (!final);

  return 1;
}
